import tensornetwork as tn
from ttnopt.src.PhysicsEngine import PhysicsEngine


class TimeEvolution(PhysicsEngine):
    def __init__(
        self,
        psi,
        physical_spin_nums,
        hamiltonians,
        max_bond_dim=100,
        max_truncation_err=1e-11,
    ):
        super().__init__(
            psi, physical_spin_nums, hamiltonians, max_bond_dim, max_truncation_err
        )

    def run(
        self,
        dt,
        opt_structure=False,
    ):
        while self.candidate_edge_ids() != []:
            (
                edge_id,
                selected_tensor_id,
                connected_tensor_id,
                not_selected_tensor_id,
            ) = self.local_two_tensor()

            # absorb gauge tensor
            iso = tn.Node(self.psi.tensors[selected_tensor_id])
            gauge = tn.Node(self.psi.gauge_tensor)
            iso[2] ^ gauge[0]
            iso = tn.contractors.auto(
                [iso, gauge], output_edge_order=[iso[0], iso[1], gauge[1]]
            )
            self.psi.tensors[selected_tensor_id] = iso.get_tensor()

            self.set_flag(not_selected_tensor_id)

            self.set_ttn_properties_at_one_tensor(edge_id, selected_tensor_id)

            self._set_edge_spin(not_selected_tensor_id)
            self._set_block_hamiltonian(not_selected_tensor_id)

            psi = self.lanczos_exp_multiply(
                [selected_tensor_id, connected_tensor_id], dt
            )
            psi_edges = (
                self.psi.edges[selected_tensor_id][:2]
                + self.psi.edges[connected_tensor_id][:2]
            )

            u, s, v, edge_order = self.decompose_two_tensors(
                psi, opt_structure=opt_structure, operate_degeneracy=True
            )

            self.psi.tensors[selected_tensor_id] = u
            self.psi.tensors[connected_tensor_id] = v
            self.psi.gauge_tensor = s
            (
                self.psi.edges[selected_tensor_id][0],
                self.psi.edges[selected_tensor_id][1],
            ) = (
                psi_edges[edge_order[0]],
                psi_edges[edge_order[1]],
            )
            (
                self.psi.edges[connected_tensor_id][0],
                self.psi.edges[connected_tensor_id][1],
            ) = (
                psi_edges[edge_order[2]],
                psi_edges[edge_order[3]],
            )

            self.distance = self.initial_distance()



from setuptools import find_packages, setup

setup(
    name="ttnopt",
    version="0.1.0",
    description="A Python package for tree tensor network algorithms",
    author="Ryo Watanabe, Hidetaka Manabe",
    author_email="manabe@acs.i.kyoto-u.ac.jp",
    url="https://github.com/Watayo/TTNOpt",
    packages=find_packages(),
    zip_safe=False,
    python_requires=">=3.6",
    license="Apache-2.0",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
    ],
    entry_points={
        "console_scripts": [
            "gss=ttnopt:ground_state_search",  # Link ttnopt_gss command to your main function
            "ft=ttnopt:factorize_tensor",  # Link ttnopt_gss command to your main function
            "samplettn=ttnopt:sample",  # Link ttnopt_sample command to your sample function
        ],
    },
    install_requires=[
        "numpy>=1.23.5",
        "jaxlib<0.4.34",
        "tensornetwork",
        "networkx",
        "pydot",
        "pyyaml",
        "dotmap",
        "matplotlib",
        "scipy",
        "pandas",
    ],
)



from .Observable import Observable
from .Hamiltonian import Hamiltonian
from .TTN import TreeTensorNetwork
from .GroundStateSearch import GroundStateSearch
from .GroundStateSearchSparse import GroundStateSearchSparse
from .FactorizeTensor import FactorizeTensor



from ttnopt.src import Hamiltonian
from ttnopt.src import TreeTensorNetwork
from ttnopt.src import GroundStateSearch

import math
import numpy as np


def open_adjacent_indexs(d: int):
    n = 2**d
    if n > 2:
        ind_list = [[i, (i + 1)] for i in range(n - 1)]
    else:
        ind_list = [[0, 1]]
    return ind_list


def heisenberg_chain_hamiltonian(d, coef_j=1.0):
    spin_size = ["S=1/2" for i in range(d**2)]
    indices = open_adjacent_indexs(d)
    interaction_coefs = [[coef_j, coef_j] for _ in indices]
    model = "XXZ"
    return Hamiltonian(
        d**2, spin_size, model, indices, interaction_coefs=interaction_coefs
    )


def test_run_dmrg():
    d = 4
    psi = TreeTensorNetwork.mps(d**2)
    hamiltonian = heisenberg_chain_hamiltonian(d)
    init_bond_dim = 4
    max_bond_dim = 10
    gss = GroundStateSearch(
        psi=psi,
        hamiltonian=hamiltonian,
        init_bond_dim=init_bond_dim,
        max_bond_dim=max_bond_dim,
    )
    gss.run(opt_structure=True, max_num_sweep=10)
    energy = gss.energy
    energy = np.mean(list(energy.values()))
    assert math.isclose(energy, -6.911621, abs_tol=1e-5)



from pathlib import Path

import pandas as pd
from dotmap import DotMap

from ttnopt.src import Hamiltonian


def hamiltonian(config: DotMap):
    """_summary_

    Args:e
        config Dict: the configuration of the system
    """
    if isinstance(config.spin_size, DotMap) and isinstance(
        config.spin_size.file, DotMap
    ):
        print("=" * 50)
        print("⚠️  Error: Please input spin value in spin_size as S or Si_file")
        print("=" * 50)
        exit()

    spin_sizes = None
    if isinstance(config.spin_size, str):
        if Path(config.spin_size).suffix == ".dat":
            spin_csv = pd.read_csv(config.spin_size, delimiter=",", header=None)
            spin_csv_sorted = spin_csv.sort_values(by=0)
            spin_size = spin_csv_sorted.iloc[:, 1].values
            spin_sizes = [f"S={s}" for s in spin_size]
        else:
            spin_sizes = ["S=" + str(config.spin_size)] * config.N
    else:
        print("=" * 50)
        print(
            "⚠️ Error : spin_size is not specified correctly either number or .dat file"
        )
        print("=" * 50)
        exit()

    interaction_indices = None
    interaction_coefs = None
    if Path(config.model.file).suffix == ".dat":
        interaction_csv = pd.read_csv(config.model.file, delimiter=",", header=None)
        interaction_indices = interaction_csv.iloc[:, :2].values
        interaction_coefs = interaction_csv.iloc[:, 2:].values
    else:
        print("=" * 50)
        print("⚠️ Error : model.file is not specified correctly")
        print("=" * 50)
        exit()

    if interaction_indices.max() >= config.N:
        print("=" * 50)
        print(
            f"⚠️  Error: XXY, XYZ interaction indices exceed the allowed range. All indices must be less than N-1 (N={config.system.N})."
        )
        print("=" * 50)
        exit()

    # magnetic_field_X
    magnetic_field_X_indices = None
    magnetic_field_X = None
    if not isinstance(config.MF_X, DotMap):
        if Path(str(config.MF_X)).suffix == ".dat":
            magnetic_field_X_csv = pd.read_csv(config.MF_X, delimiter=",", header=None)
            magnetic_field_X_indices = magnetic_field_X_csv.iloc[:, 0].values
            magnetic_field_X = magnetic_field_X_csv.iloc[:, 1].values
        else:
            if isinstance(float(config.MF_X), float):
                magnetic_field_X_indices = [i for i in range(config.N)]
                magnetic_field_X = [config.MF_X] * config.N
            else:
                print("=" * 50)
                print("⚠️  Error: MF_X should be .dat file or float.")
                print("=" * 50)
                exit()

    # magnetic_field_Y
    magnetic_field_Y_indices = None
    magnetic_field_Y = None
    if not isinstance(config.MF_Y, DotMap):
        if Path(str(config.MF_Y)).suffix == ".dat":
            magnetic_field_Y_csv = pd.read_csv(config.MF_Y, delimiter=",", header=None)
            magnetic_field_Y_indices = magnetic_field_Y_csv.iloc[:, 0].values
            magnetic_field_Y = magnetic_field_Y_csv.iloc[:, 1].values
        else:
            if isinstance(float(config.MF_Y), float):
                magnetic_field_Y_indices = [i for i in range(config.N)]
                magnetic_field_Y = [config.MF_Y] * config.N
            else:
                print("=" * 50)
                print("⚠️  Error: MF_Y should be .dat file or float.")
                print("=" * 50)
                exit()

    # magnetic_field_Z
    magnetic_field_Z_indices = None
    magnetic_field_Z = None
    if not isinstance(config.MF_Z, DotMap):
        if Path(str(config.MF_Z)).suffix == ".dat":
            magnetic_field_Z_csv = pd.read_csv(config.MF_Z, delimiter=",", header=None)
            magnetic_field_Z_indices = magnetic_field_Z_csv.iloc[:, 0].values
            magnetic_field_Z = magnetic_field_Z_csv.iloc[:, 1].values
        else:
            if isinstance(float(config.MF_Z), float):
                magnetic_field_Z_indices = [i for i in range(config.N)]
                magnetic_field_Z = [config.MF_Z] * config.N
            else:
                print("=" * 50)
                print("⚠️  Error: MF_Z should be .dat file or float.")
                print("=" * 50)
                exit()

    ion_anisotropy_indices = None
    ion_anisotropy = None
    if not isinstance(config.SIA, DotMap):
        if Path(str(config.SIA)).suffix == ".dat":
            ion_anisotropy_csv = pd.read_csv(config.SIA, delimiter=",", header=None)
            ion_anisotropy_indices = ion_anisotropy_csv.iloc[:, 0].values
            ion_anisotropy = ion_anisotropy_csv.iloc[:, 1].values
        else:
            if isinstance(float(config.SIA), float):
                ion_anisotropy_indices = [i for i in range(config.N)]
                ion_anisotropy = [config.SIA] * config.N
            else:
                print("=" * 50)
                print("⚠️  Error: SIA should be .dat file or float.")
                print("=" * 50)
                exit()

    # dzyaloshinskii_moriya_X
    dzyaloshinskii_moriya_X_indices = None
    dzyaloshinskii_moriya_X = None
    if not isinstance(config.DM_X, DotMap):
        if Path(config.DM_X).suffix == ".dat":
            dzyaloshinskii_moriya_X_csv = pd.read_csv(
                config.DM_X, delimiter=",", header=None
            )
            dzyaloshinskii_moriya_X_indices = dzyaloshinskii_moriya_X_csv.iloc[
                :, :2
            ].values
            dzyaloshinskii_moriya_X = dzyaloshinskii_moriya_X_csv.iloc[:, 2:].values
            if dzyaloshinskii_moriya_X_indices.max() >= config.N:
                print("=" * 50)
                print(
                    f"⚠️  Error: DM_X interaction indices exceed the allowed range. All indices must be less than N-1 (N={config.system.N})."
                )
                print("=" * 50)
                exit()
        else:
            print("=" * 50)
            print("⚠️  Error: DM_X should be .dat file.")
            print("=" * 50)
            exit()

    # dzyaloshinskii_moriya_Y
    dzyaloshinskii_moriya_Y_indices = None
    dzyaloshinskii_moriya_Y = None
    if not isinstance(config.DM_Y, DotMap):
        if Path(config.DM_Y).suffix == ".dat":
            dzyaloshinskii_moriya_Y_csv = pd.read_csv(
                config.DM_Y, delimiter=",", header=None
            )
            dzyaloshinskii_moriya_Y_indices = dzyaloshinskii_moriya_Y_csv.iloc[
                :, :2
            ].values
            dzyaloshinskii_moriya_Y = dzyaloshinskii_moriya_Y_csv.iloc[:, 2:].values
            if dzyaloshinskii_moriya_Y_indices.max() >= config.N:
                print("=" * 50)
                print(
                    f"⚠️  Error: DM_Y interaction indices exceed the allowed range. All indices must be less than N-1 (N={config.system.N})."
                )
                print("=" * 50)
                exit()
        else:
            print("=" * 50)
            print("⚠️  Error: DM_Y should be .dat file.")
            print("=" * 50)
            exit()

    # dzyaloshinskii_moriya_Z
    dzyaloshinskii_moriya_Z_indices = None
    dzyaloshinskii_moriya_Z = None
    if not isinstance(config.DM_Z, DotMap):
        if Path(config.DM_Z).suffix == ".dat":
            dzyaloshinskii_moriya_Z_csv = pd.read_csv(
                config.DM_Z, delimiter=",", header=None
            )
            dzyaloshinskii_moriya_Z_indices = dzyaloshinskii_moriya_Z_csv.iloc[
                :, :2
            ].values
            dzyaloshinskii_moriya_Z = dzyaloshinskii_moriya_Z_csv.iloc[:, 2:].values
            if dzyaloshinskii_moriya_Z_indices.max() >= config.N:
                print("=" * 50)
                print(
                    f"⚠️  Error: DM_Z interaction indices exceed the allowed range. All indices must be less than N-1 (N={config.system.N})."
                )
                print("=" * 50)
                exit()
        else:
            print("=" * 50)
            print("⚠️  Error: DM_Y should be .dat file.")
            print("=" * 50)
            exit()

    # sod_x
    sod_X_indices = None
    sod_X = None
    if not isinstance(config.SOD_X, DotMap):
        if Path(config.SOD_X).suffix == ".dat":
            sod_x_csv = pd.read_csv(config.SOD_X, delimiter=",", header=None)
            sod_X_indices = sod_x_csv.iloc[:, :2].values
            sod_X = sod_x_csv.iloc[:, 2:].values
            if sod_X_indices.max() >= config.N:
                print("=" * 50)
                print(
                    f"⚠️  Error: SOD_X interaction indices exceed the allowed range. All indices must be less than N-1 (N={config.system.N})."
                )
                print("=" * 50)
                exit()

    # sod_y
    sod_Y_indices = None
    sod_Y = None
    if not isinstance(config.SOD_Y, DotMap):
        if Path(config.SOD_Y).suffix == ".dat":
            sod_y_csv = pd.read_csv(config.SOD_Y, delimiter=",", header=None)
            sod_Y_indices = sod_y_csv.iloc[:, :2].values
            sod_Y = sod_y_csv.iloc[:, 2:].values
            if sod_Y_indices.max() >= config.N:
                print("=" * 50)
                print(
                    f"⚠️  Error: SOD_Y interaction indices exceed the allowed range. All indices must be less than N-1 (N={config.system.N})."
                )
                print("=" * 50)
                exit()

    # sod_Z
    sod_Z_indices = None
    sod_Z = None
    if not isinstance(config.SOD_Z, DotMap):
        if Path(config.SOD_Z).suffix == ".dat":
            sod_z_csv = pd.read_csv(config.SOD_Z, delimiter=",", header=None)
            sod_Z_indices = sod_z_csv.iloc[:, :2].values
            sod_Z = sod_z_csv.iloc[:, 2:].values
            if sod_Z_indices.max() >= config.N:
                print("=" * 50)
                print(
                    f"⚠️  Error: SOD_Z interaction indices exceed the allowed range. All indices must be less than N-1 (N={config.system.N})."
                )
                print("=" * 50)
                exit()

    if config.model.type == "XXZ":
        if interaction_coefs.shape[1] == 2:
            hamiltonian = Hamiltonian(
                config.N,
                spin_sizes,
                config.model.type,
                interaction_indices,
                interaction_coefs,
                magnetic_field_X_indices=magnetic_field_X_indices,
                magnetic_field_X=magnetic_field_X,
                magnetic_field_Y_indices=magnetic_field_Y_indices,
                magnetic_field_Y=magnetic_field_Y,
                magnetic_field_Z_indices=magnetic_field_Z_indices,
                magnetic_field_Z=magnetic_field_Z,
                ion_anisotropy_indices=ion_anisotropy_indices,
                ion_anisotropy=ion_anisotropy,
                dzyaloshinskii_moriya_X_indices=dzyaloshinskii_moriya_X_indices,
                dzyaloshinskii_moriya_X=dzyaloshinskii_moriya_X,
                dzyaloshinskii_moriya_Y_indices=dzyaloshinskii_moriya_Y_indices,
                dzyaloshinskii_moriya_Y=dzyaloshinskii_moriya_Y,
                dzyaloshinskii_moriya_Z_indices=dzyaloshinskii_moriya_Z_indices,
                dzyaloshinskii_moriya_Z=dzyaloshinskii_moriya_Z,
                sod_X_indices=sod_X_indices,
                sod_X=sod_X,
                sod_Y_indices=sod_Y_indices,
                sod_Y=sod_Y,
                sod_Z_indices=sod_Z_indices,
                sod_Z=sod_Z,
            )
        else:
            print("=" * 50)
            print("⚠️  Error: Please input two columns in model.file for XXZ model")
            print("=" * 50)
            exit()
    elif config.model.type == "XYZ":
        if interaction_coefs.shape[1] == 3:
            hamiltonian = Hamiltonian(
                config.N,
                spin_sizes,
                config.model.type,
                interaction_indices,
                interaction_coefs,
                magnetic_field_X_indices=magnetic_field_X_indices,
                magnetic_field_X=magnetic_field_X,
                magnetic_field_Y_indices=magnetic_field_Y_indices,
                magnetic_field_Y=magnetic_field_Y,
                magnetic_field_Z_indices=magnetic_field_Z_indices,
                magnetic_field_Z=magnetic_field_Z,
                ion_anisotropy_indices=ion_anisotropy_indices,
                ion_anisotropy=ion_anisotropy,
                dzyaloshinskii_moriya_X_indices=dzyaloshinskii_moriya_X_indices,
                dzyaloshinskii_moriya_X=dzyaloshinskii_moriya_X,
                dzyaloshinskii_moriya_Y_indices=dzyaloshinskii_moriya_Y_indices,
                dzyaloshinskii_moriya_Y=dzyaloshinskii_moriya_Y,
                dzyaloshinskii_moriya_Z_indices=dzyaloshinskii_moriya_Z_indices,
                dzyaloshinskii_moriya_Z=dzyaloshinskii_moriya_Z,
                sod_X_indices=sod_X_indices,
                sod_X=sod_X,
                sod_Y_indices=sod_Y_indices,
                sod_Y=sod_Y,
                sod_Z_indices=sod_Z_indices,
                sod_Z=sod_Z,
            )
        else:
            print("=" * 50)
            print("⚠️  Error: Please input three columns in model.file for XYZ model")
            print("=" * 50)
            exit()
    else:
        print("=" * 50)
        print("⚠️  Error: Please input the correct model type (XXZ or XYZ)")
        print("=" * 50)
        exit()

    return hamiltonian



from .src.Observable import Observable
from .src.Hamiltonian import Hamiltonian
from .src.TTN import TreeTensorNetwork
from. src.GroundStateSearch import GroundStateSearch
from .src.GroundStateSearchSparse import GroundStateSearchSparse

from .hamiltonian import hamiltonian
from .ground_state_search import ground_state_search
from .factorize_tensor import factorize_tensor